// TODO

// WARNING: This file has automatically been generated on Wed, 12 Feb 2020 10:38:25 UTC.
// Code generated by https://git.io/c-for-go. DO NOT EDIT.

/*
Package nll2tp provides Go bindings for the Linux L2TP GeNetlink API
*/
package nll2tp
